#include <cstdio>
#include <string>
#include <vector>
#include <cstring>

#include <omp.h>

using namespace std ;


#define BUCKET_BASES    5
#define BUCKET_NUM      (1UL<<(BUCKET_BASES<<1))


size_t base_to_bit_lookup[128] ;
char fix_base_lookup[128] ;
char complement_base_lookup[128] ;


size_t bucket_id( const char *sequence ){
    size_t re = 0 ;
    for( size_t i=0 ; i<BUCKET_BASES ; i++ ){
        re <<= 2 ;
        re |= base_to_bit_lookup[(size_t)sequence[i]] ;
    }
    return re ;
}


// complement sequence
void complement_sequence( const char *seq, char *re, size_t len = 0 ){
    if( !len )
        len = strlen( seq ) ;
    else
        seq += strlen(seq) - len ;

    const size_t end = len-1 ;

    for( size_t i=0 ; i<=end ; i++ )
        re[end-i] = complement_base_lookup[(size_t)seq[i]] ;

    re[end+1] = '\0' ;
}


// basic type of a read set
class ReadSet{
public:
    char *sequence ;
    bool is_forward ;

    size_t ID ;
    ReadSet *complement ;
} ;


// compare read set
int cmp_sequence( const ReadSet *a, const ReadSet *b ){
    const char *str_a = a->sequence ;
    const char *str_b = b->sequence ;

    // both are forward
    if( a->is_forward && b->is_forward )
        return strcmp( str_a, str_b ) ;

    else{
        size_t i_a, i_b ;

        // both are reversed
        if( !a->is_forward && !b->is_forward ){
            i_a = strlen( str_a ) ;
            i_b = strlen( str_b ) ;

            while( i_a && i_b ){
                if( str_a[--i_a] != str_b[--i_b] )
                    return str_b[i_b] - str_a[i_a] ;
            }
        }

        // b is reversed
        else if( a->is_forward ){
            i_a = 0 ;
            i_b = strlen( str_b ) ;

            char base_a ;
            while( (base_a=str_a[i_a++]) && i_b ){
                char base_b = complement_base_lookup[(size_t)str_b[--i_b]] ;
                if( base_a != base_b )
                    return base_a - base_b ;
            }

            if( !base_a )
                i_a = 0 ;
        }

        // a is reversed
        else{
            i_a = strlen( str_a ) ;
            i_b = 0 ;

            char base_b ;
            while( (base_b=str_b[i_b++]) && i_a ){
                char base_a = complement_base_lookup[(size_t)str_a[--i_a]] ;
                if( base_a != base_b )
                    return base_a - base_b ;
            }

            if( !base_b )
                i_b = 0 ;
        }

        return (int)i_a - (int)i_b ;
    }

    return 0 ;
}


// sorter of reads
class ReadSorter{
    vector <ReadSet*> read ;
    bool dirty ;

    vector <size_t> pivot_loc ;
    vector <size_t> read_count ;

    void heap_moving( size_t index, size_t size ){
        ReadSet *pivot = read[index] ;
        
        size_t max_child ;
        while( (max_child = (index<<1)+1) < size ){
            // find maximum child
            if( max_child+1 < size ){
                if( cmp_sequence(read[max_child+1], read[max_child]) > 0 )
                    max_child ++ ;
            }

            // check max-heap
            if( cmp_sequence(pivot, read[max_child]) >= 0 )
                break ;

            // swap
            read[index] = read[max_child] ;
            index = max_child ;
        }

        read[index] = pivot ;
    }

public:
    ReadSorter(){
        dirty = false ;
    }
    ~ReadSorter(){
        for( size_t i=0 ; i<read.size() ; i++ ){
            if( read[i]->is_forward )
                delete [] read[i]->sequence ;
            delete read[i] ;
        }
    }

    // append entry
    void push_back( ReadSet* const entry ){
        read.push_back( entry ) ;
        dirty = true ;
    }

    // sort
    void sort(){
        if( !dirty )
            return ;

        // heapify
        size_t size = read.size() ;
        for( size_t index=size>>1 ; index-- ; )
            heap_moving( index, size ) ;

        // heap sort
        while( size > 1 ){
            --size ;
            ReadSet *back = read[size] ;
            read[size] = read[0] ;
            read[0] = back ;

            heap_moving( 0, size ) ;
        }

        dirty = false ;
    }

    // label ID
    size_t label_ID( size_t ID_bias ){
        // init
        ReadSet *pivot = NULL ;
        size_t set_count = 0 ;

        // scan
        for( size_t i=0 ; i<read.size() ; i++ ){
            if( pivot==NULL || cmp_sequence(pivot, read[i]) ){
                pivot = read[i] ;
                pivot->ID = ID_bias + set_count ;

                pivot_loc.push_back( i ) ;
                read_count.push_back( 0 ) ;
                set_count ++ ;
            }

            read[i]->ID = pivot->ID ;

            if( read[i]->is_forward ){
                pivot_loc.back() = i ;
                read_count.back() ++ ;
            }
        }

        return set_count ;
    }

    // print
    void print(){
        size_t set_count = 0 ;
        char str_buf[1024] ;

        // scan
        for( size_t i=0 ; i<pivot_loc.size() ; i++ ){
            const ReadSet &entry = *read[pivot_loc[i]] ;

            if( !entry.is_forward )
                complement_sequence( entry.sequence, str_buf ) ;

            // read count
            fprintf( stdout, "%lu\t%s\t%lu\n", 
                read_count[set_count++],
                entry.is_forward?
                    entry.sequence:
                    str_buf,
                entry.complement->ID ) ;
        }
    }
} ;


void init_lookup(){
    for( size_t i=0 ; i<128 ; i++ ){
        base_to_bit_lookup[i] = (
            i=='C'?1:
            i=='G'?2:
            i=='T'?3:
            0 ) ;

        complement_base_lookup[i] = (
            i=='C'?'G':
            i=='G'?'C':
            i=='T'?'A':
            'T' ) ;

        fix_base_lookup[i] = (
            i=='\n'?'\0':
            i=='C'?'C':
            i=='c'?'C':
            i=='G'?'G':
            i=='g'?'G':
            i=='T'?'T':
            i=='t'?'T':
            'A' ) ;
    }
}


int main(int argc, char *argv[]){
    size_t readLen = 0 ;
    size_t totalNode = 0 ;

    ReadSorter sorter[BUCKET_NUM] ;

    // initialize lookup table
    init_lookup() ;

    // parse FASTA
    char line[1024] ;
    string sequence ;
    do{
        // end of a sequence
        if( !fgets(line, sizeof(line), stdin) || line[0] == '>' ){
            if( !sequence.size() )
                continue ;

            // generate reads
            // palindrome is duplicated and safe.
            ReadSet *read_f = new ReadSet ;
            ReadSet *read_r = new ReadSet ;

            read_f->sequence = read_r->sequence = new char[sequence.size()+1] ;
            strcpy( read_f->sequence, sequence.c_str() ) ;

            read_f->is_forward = true ;
            read_r->is_forward = false ;

            read_f->complement = read_r ;
            read_r->complement = read_f ;

            // append
            sorter[bucket_id(sequence.c_str())].push_back( read_f ) ;
            complement_sequence( sequence.c_str(), line, BUCKET_BASES ) ;
            sorter[bucket_id(line)].push_back( read_r ) ;            

            // update
            if( readLen < sequence.size() )
                readLen = sequence.size() ;
            sequence.clear() ;
        }

        // append sequence
        else{
            // fix all bases
            for( size_t i=0 ; line[i] ; i++ )
                line[i] = fix_base_lookup[(size_t)line[i]] ;
            sequence += line ;
        }
    } while( !feof(stdin) ) ;
    
    // sort
    #pragma omp parallel for schedule(dynamic)
    for( size_t i=0 ; i<BUCKET_NUM ; i++ )
        sorter[i].sort() ;

    // label ID
    for( size_t i=0 ; i<BUCKET_NUM ; i++ )
        totalNode += sorter[i].label_ID( totalNode ) ;

    // output
    fprintf( stdout, "readNum= %lu\t readLength= %lu\n", totalNode, readLen ) ;
    for( size_t i=0 ; i<BUCKET_NUM ; i++ )
        sorter[i].print() ;

    return 0 ;
}



